/*
    Robot server - 2005 Jason Hunt
    
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com
    
    File: Camera_drv.c
    
    
    Adapted from V4L2 reference code
*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <asm/types.h>          /* for videodev2.h */

#include <linux/videodev2.h>

#define CLEAR(x) memset (&(x), 0, sizeof (x))

#include "camera_drv.h"



/*******************************************************************************#
#  huffman tables for Jpeg encoder/decoder                                      #
#           Paulo Assis <pj.assis@gmail.com>                                    #
********************************************************************************/


#define JPG_HUFFMAN_TABLE_LENGTH 0x01A0

static const unsigned char JPEGHuffmanTable[JPG_HUFFMAN_TABLE_LENGTH] = 
{
	// luminance dc - length bits	
	0x00, 
	0x00, 0x01, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	// luminance dc - code
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09,
	0x0A, 0x0B, 
	// chrominance dc - length bits	
	0x01, 
	0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
	// chrominance dc - code
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09,
	0x0A, 0x0B, 
	// luminance ac - number of codes with # bits (ordered by code length 1-16)
	0x10,
	0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03, 0x05, 0x05, 
	0x04, 0x04, 0x00, 0x00, 0x01, 0x7D,
	// luminance ac - run size (ordered by code length)	
	0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12, 0x21, 0x31,
	0x41, 0x06, 0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32,
	0x81, 0x91, 0xA1, 0x08, 0x23, 0x42, 0xB1, 0xC1, 0x15, 0x52,
	0xD1, 0xF0, 0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0A, 0x16,
	0x17, 0x18, 0x19, 0x1A, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A,
	0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x43, 0x44, 0x45,
	0x46, 0x47, 0x48, 0x49, 0x4A, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5A, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
	0x6A, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A, 0x83,
	0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x92, 0x93, 0x94,
	0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5,
	0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6,
	0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7,
	0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8,
	0xD9, 0xDA, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8,
	0xE9, 0xEA, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8,
	0xF9, 0xFA, 
	// chrominance ac -number of codes with # bits (ordered by code length 1-16)
	0x11, 
	0x00, 0x02, 0x01, 0x02, 0x04, 0x04, 0x03, 0x04, 0x07, 0x05,
	0x04, 0x04, 0x00, 0x01, 0x02, 0x77,
	// chrominance ac - run size (ordered by code length)
	0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21, 0x31, 0x06,
	0x12, 0x41, 0x51, 0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81,
	0x08, 0x14, 0x42, 0x91, 0xA1, 0xB1, 0xC1, 0x09, 0x23, 0x33,
	0x52, 0xF0, 0x15, 0x62, 0x72, 0xD1, 0x0A, 0x16, 0x24, 0x34,
	0xE1, 0x25, 0xF1, 0x17, 0x18, 0x19, 0x1A, 0x26, 0x27, 0x28,
	0x29, 0x2A, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x43, 0x44,
	0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x53, 0x54, 0x55, 0x56,
	0x57, 0x58, 0x59, 0x5A, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
	0x69, 0x6A, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7A,
	0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x92,
	0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0xA2, 0xA3,
	0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4,
	0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5,
	0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6,
	0xD7, 0xD8, 0xD9, 0xDA, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7,
	0xE8, 0xE9, 0xEA, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8,
	0xF9, 0xFA
};



typedef enum 
{
	IO_METHOD_READ,
	IO_METHOD_MMAP,
	IO_METHOD_USERPTR,
} io_method;

struct buffer 
{
        void *                  start;
        size_t                  length;
};

char   dev_name[256];



io_method	io		= IO_METHOD_MMAP;


int fd = -1;
struct buffer *buffers = NULL;
unsigned int n_buffers = 0;


unsigned int image_x = 0;
unsigned int image_y = 0;


unsigned int max_image = 0;




void errno_exit                      (const char *           s)
{
        fprintf (stderr, "%s error %d, %s\n",
                 s, errno, strerror (errno));

        exit (EXIT_FAILURE);
}

int xioctl                          (int                    fd,
                                 int                    request,
                                 void *                 arg)
{
        int r;

        do r = ioctl (fd, request, arg);
        while (-1 == r && EINTR == errno);

        return r;
}







int clamp (double x)
{
	int r = x;      /* round to nearest */

	if (r < 0)         return 0;
	else if (r > 255)  return 255;
	else               return r;
}



/* Insert a default huffman table. */
// Code adapted from guvcview
static void process_image(unsigned char * image_in, unsigned char *image_out, unsigned int in_len, unsigned int *out_len)
{

	int ret=0;
	int jpgsize=0;
//	unsigned char  *jpgtmp=NULL;
	unsigned char  *Pimg=NULL;
	unsigned char  *Pjpg=NULL;
	unsigned char  *tp=NULL;	
	JPGFILEHEADER JpgFileh;
	int jpghsize=sizeof(JpgFileh);
	FILE *fp;
	
	
	jpgsize=jpghsize+in_len+sizeof(JPEGHuffmanTable)+4;/*header+huffman+marker+buffsize*/


	Pimg=image_in;
	
//	if((jpgtmp=g_malloc0(jpgsize))!=NULL) 
	
	
	
	
//	{
		Pjpg=image_out;
		/*Fill JFIF header*/
		JpgFileh.SOI[0]=0xff;
		JpgFileh.SOI[1]=0xd8;
		JpgFileh.APP0[0]=0xff;
		JpgFileh.APP0[1]=0xe0;
		JpgFileh.length[0]=0x00;
		JpgFileh.length[1]=0x10;
		JpgFileh.JFIF[0]=0x4a;//JFIF0
		JpgFileh.JFIF[1]=0x46;
		JpgFileh.JFIF[2]=0x49;
		JpgFileh.JFIF[3]=0x46;
		JpgFileh.JFIF[4]=0x00;
		JpgFileh.VERS[0]=0x01;//version 1.2
		JpgFileh.VERS[1]=0x02;
		JpgFileh.density=0x00;
		JpgFileh.xdensity[0]=0x00;
		JpgFileh.xdensity[1]=0x78;
		JpgFileh.ydensity[0]=0x00;
		JpgFileh.ydensity[1]=0x78;
		JpgFileh.WTN=0;
		JpgFileh.HTN=0;
	
		/*adds header (JFIF)*/
		memmove(Pjpg,&JpgFileh,jpghsize);
		/*moves to the end of the header struct (JFIF)*/
		Pjpg+=jpghsize;
		int headSize = image_in[4]*256+image_in[5] + 4;/*length + SOI+APP0*/
		/*moves to the end of header (MJPG)*/
		Pimg+=headSize;
		/*adds Quantization tables and everything else until   * 
		* start of frame marker (FFC0) 	 */
		tp=Pimg;
		int qtsize=0;
		while(!((tp[qtsize]== 0xff) && (tp[qtsize+1]== 0xc0))) 
		{
			qtsize++;
		}
		memmove(Pjpg,Pimg,qtsize);
		/*moves to the begining of frame marker*/
		Pjpg+=qtsize; 
		Pimg+=qtsize;
		/*insert huffman table with marker (FFC4) and length(x01a2)*/
		BYTE HUFMARK[4];
		HUFMARK[0]=0xff;
		HUFMARK[1]=0xc4;
		HUFMARK[2]=0x01;
		HUFMARK[3]=0xa2;
		memmove(Pjpg,&HUFMARK,4);
		Pjpg+=4;
		memmove(Pjpg,&JPEGHuffmanTable,JPG_HUFFMAN_TABLE_LENGTH);/*0x01a0*/
		/*moves to the end of huffman tables (JFIF)*/
		Pjpg+=JPG_HUFFMAN_TABLE_LENGTH;
		/*copys frame data(JFIF)*/
		memmove(Pjpg,Pimg,(in_len-(Pimg-image_in)));
		Pjpg+=in_len-(Pimg-image_in);
		
		int totSize = Pjpg - image_out;
	
	
	
	/*	if ((fp = fopen(Filename,"wb"))!=NULL) 
		{
			ret=fwrite(image_out,totSize,1,fp);  //jpeg - jfif
			if (ret< 1) ret=1; //write error 
			else ret=0;
			fflush(fp); //flush data stream to file system
			if(fsync(fileno(fp)) || fclose(fp))
				perror("JPEG ERROR - couldn't write to file");
		} 
		else ret=1;*/

//		g_free(image_out);
//		jpgtmp=NULL;
		Pimg=NULL;
		Pjpg=NULL;
		tp=NULL;
//	}

    *out_len = totSize ;
    
    
    //memcpy(out, in, in_len);



}













int
read_frame		    (unsigned char *data, unsigned int max, unsigned int *len)
{

           struct v4l2_buffer buf;
	unsigned int i;

	switch (io) {
	case IO_METHOD_READ:
    		if (-1 == read (fd, buffers[0].start, buffers[0].length)) {
            		switch (errno) {
            		case EAGAIN:
                    		return 0;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				errno_exit ("read");
			}
		}

       // *len = buffers[0].length;
    	//	process_image (buffers[0].start, data, max, *len);
    		
   			errno_exit ("IO_METHOD_READ Not supported");

    		
//    		memcpy(data, buffers[0].start, buffers[0].length);
//    		*len = buffers[0].length;
    		
    		

		break;

	case IO_METHOD_MMAP:
		CLEAR (buf);

            	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            	buf.memory = V4L2_MEMORY_MMAP;

    		if (-1 == xioctl (fd, VIDIOC_DQBUF, &buf)) {
            		switch (errno) {
            		case EAGAIN:
                    		return 0;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				errno_exit ("VIDIOC_DQBUF");
			}
		}

                assert (buf.index < n_buffers);

        //*len = buffers[0].length;

    process_image(buffers[buf.index].start, data, buf.bytesused, len);
    

//        	*len = buf.bytesused;
//    		memcpy(data, buffers[buf.index].start, buf.bytesused);
    	



		if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
			errno_exit ("VIDIOC_QBUF");

		break;

	case IO_METHOD_USERPTR:
		CLEAR (buf);

    		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    		buf.memory = V4L2_MEMORY_USERPTR;

		if (-1 == xioctl (fd, VIDIOC_DQBUF, &buf)) {
			switch (errno) {
			case EAGAIN:
				return 0;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				errno_exit ("VIDIOC_DQBUF");
			}
		}

		for (i = 0; i < n_buffers; ++i)
			if (buf.m.userptr == (unsigned long) buffers[i].start
			    && buf.length == buffers[i].length)
				break;

		assert (i < n_buffers);

	errno_exit ("IO_METHOD_USERPTR Not supported");

 //       	*len = buf.length;
 //   		process_image ((void *) buf.m.userptr, data, max, *len);
    		

//  			memcpy(data, (void *) buf.m.userptr, buf.length);
//    		*len = buf.length;



		if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
			errno_exit ("VIDIOC_QBUF");

		break;
	}

	return 1;
}


void cameradrv_frame (unsigned char *data, unsigned int max, unsigned int *len)
{


//printf("camera frame\n");

                for (;;) 
                {
                        fd_set fds;
                        struct timeval tv;
                        int r;

                        FD_ZERO (&fds);
                        FD_SET (fd, &fds);


//printf("After fd set\n");




                        /* Timeout. */
                        tv.tv_sec = 2;
                        tv.tv_usec = 0;
//printf("before select\b");

             
                         r = select (fd + 1, &fds, NULL, NULL, &tv);
//printf("after selec\b");

                        if (-1 == r) {
                                if (EINTR == errno)
                                        continue;

                                errno_exit ("select");
                        }

                        if (0 == r) {
                                fprintf (stderr, "select timeout\n");
                                exit (EXIT_FAILURE);
                        }


//printf("before read\b");

			if (read_frame (data, max, len))
                    		break;
//printf("after read\b");
	
			/* EAGAIN - continue select loop. */
                }

}








void stop_capturing (void)
{
    enum v4l2_buf_type type;

	switch (io) {
	case IO_METHOD_READ:
		/* Nothing to do. */
		break;

	case IO_METHOD_MMAP:
	case IO_METHOD_USERPTR:
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

		if (-1 == xioctl (fd, VIDIOC_STREAMOFF, &type))
			errno_exit ("VIDIOC_STREAMOFF");

		break;
	}
}



void start_capturing  (void)
{
        unsigned int i;
        enum v4l2_buf_type type;

	switch (io) {
	case IO_METHOD_READ:
		/* Nothing to do. */
		break;

	case IO_METHOD_MMAP:
		for (i = 0; i < n_buffers; ++i) {
            		struct v4l2_buffer buf;

        		CLEAR (buf);

        		buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        		buf.memory      = V4L2_MEMORY_MMAP;
        		buf.index       = i;

        		if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
                    		errno_exit ("VIDIOC_QBUF");
		}
		
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

		if (-1 == xioctl (fd, VIDIOC_STREAMON, &type))
			errno_exit ("VIDIOC_STREAMON");

		break;

	case IO_METHOD_USERPTR:
		for (i = 0; i < n_buffers; ++i) {
            		struct v4l2_buffer buf;

        		CLEAR (buf);

        		buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        		buf.memory      = V4L2_MEMORY_USERPTR;
			buf.index       = i;
			buf.m.userptr	= (unsigned long) buffers[i].start;
			buf.length      = buffers[i].length;

			if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
                    		errno_exit ("VIDIOC_QBUF");
		}

		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

		if (-1 == xioctl (fd, VIDIOC_STREAMON, &type))
			errno_exit ("VIDIOC_STREAMON");

		break;
	}
}

 void uninit_device  (void)
{
        unsigned int i;

	switch (io) {
	case IO_METHOD_READ:
		free (buffers[0].start);
		break;

	case IO_METHOD_MMAP:
		for (i = 0; i < n_buffers; ++i)
			if (-1 == munmap (buffers[i].start, buffers[i].length))
				errno_exit ("munmap");
		break;

	case IO_METHOD_USERPTR:
		for (i = 0; i < n_buffers; ++i)
			free (buffers[i].start);
		break;
	}

	free (buffers);
}

void  init_read			(unsigned int		buffer_size)
{
        buffers = calloc (1, sizeof (*buffers));

        if (!buffers) {
                fprintf (stderr, "Out of memory\n");
                exit (EXIT_FAILURE);
        }

	buffers[0].length = buffer_size;
	buffers[0].start = malloc (buffer_size+1);

	if (!buffers[0].start) {
    		fprintf (stderr, "Out of memory\n");
            	exit (EXIT_FAILURE);
	}
}

void init_mmap			(void)
{
	struct v4l2_requestbuffers req;

        CLEAR (req);

        req.count               = 4;
        req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory              = V4L2_MEMORY_MMAP;

	if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req)) {
                if (EINVAL == errno) {
                        fprintf (stderr, "%s does not support "
                                 "memory mapping\n", dev_name);
                        exit (EXIT_FAILURE);
                } else {
                        errno_exit ("VIDIOC_REQBUFS");
                }
        }

        if (req.count < 2) {
                fprintf (stderr, "Insufficient buffer memory on %s\n",
                         dev_name);
                exit (EXIT_FAILURE);
        }

        buffers = calloc (req.count, sizeof (*buffers));

        if (!buffers) {
                fprintf (stderr, "Out of memory\n");
                exit (EXIT_FAILURE);
        }

        for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
                struct v4l2_buffer buf;

                CLEAR (buf);

                buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory      = V4L2_MEMORY_MMAP;
                buf.index       = n_buffers;

                if (-1 == xioctl (fd, VIDIOC_QUERYBUF, &buf))
                        errno_exit ("VIDIOC_QUERYBUF");

                buffers[n_buffers].length = buf.length;
                buffers[n_buffers].start =
                        mmap (NULL /* start anywhere */,
                              buf.length,
                              PROT_READ | PROT_WRITE /* required */,
                              MAP_SHARED /* recommended */,
                              fd, buf.m.offset);

                if (MAP_FAILED == buffers[n_buffers].start)
                        errno_exit ("mmap");
        }
}
















void init_userp			(unsigned int		buffer_size)
{
	struct v4l2_requestbuffers req;
        unsigned int page_size;

        page_size = getpagesize ();
        buffer_size = (buffer_size + page_size - 1) & ~(page_size - 1);

        CLEAR (req);

        req.count               = 4;
        req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory              = V4L2_MEMORY_USERPTR;

        if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req)) {
                if (EINVAL == errno) {
                        fprintf (stderr, "%s does not support "
                                 "user pointer i/o\n", dev_name);
                        exit (EXIT_FAILURE);
                } else {
                        errno_exit ("VIDIOC_REQBUFS");
                }
        }

        buffers = calloc (4, sizeof (*buffers));

        if (!buffers) {
                fprintf (stderr, "Out of memory\n");
                exit (EXIT_FAILURE);
        }

        for (n_buffers = 0; n_buffers < 4; ++n_buffers) {
                buffers[n_buffers].length = buffer_size;
                buffers[n_buffers].start = memalign (/* boundary */ page_size,
                                                     buffer_size);

                if (!buffers[n_buffers].start) {
    			fprintf (stderr, "Out of memory\n");
            		exit (EXIT_FAILURE);
		}
        }
}







void video_try_format(char *name,  unsigned int format)
{
 struct v4l2_format fmt;
        CLEAR (fmt);
        

        fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        fmt.fmt.pix.width       = 320; 
        fmt.fmt.pix.height      = 240;
        fmt.fmt.pix.pixelformat = format ;// V4L2_PIX_FMT_YUYV;

        if (-1 == xioctl (fd, VIDIOC_TRY_FMT, &fmt)) 
        {
           // fprintf(stderr,"Format FAIL %s\n", name);
        }
        else
        {
            fprintf(stderr,"Format OK   %s\n", name);
        }
        
        
        
        

}


















 void init_device   (void)
{
        struct v4l2_capability cap;
        struct v4l2_cropcap cropcap;
        struct v4l2_crop crop;
        struct v4l2_format fmt;
	unsigned int min;

        if (-1 == xioctl (fd, VIDIOC_QUERYCAP, &cap)) {
                if (EINVAL == errno) {
                        fprintf (stderr, "%s is no V4L2 device\n",
                                 dev_name);
                        exit (EXIT_FAILURE);
                } else {
                        errno_exit ("VIDIOC_QUERYCAP");
                }
        }

        if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
                fprintf (stderr, "%s is no video capture device\n",
                         dev_name);
                exit (EXIT_FAILURE);
        }

/*

if (cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) printf("The device supports the Video Capture interface.\n");
if (cap.capabilities & V4L2_CAP_VIDEO_OUTPUT) printf("	The device supports the Video Output interface.\n");
if (cap.capabilities & V4L2_CAP_VIDEO_OVERLAY) printf("	The device supports the Video Overlay interface. A video overlay device typically stores captured images directly in the video memory of a graphics card, with hardware clipping and scaling.\n");
if (cap.capabilities & V4L2_CAP_VBI_CAPTURE	) printf("	The device supports the Raw VBI Capture interface, providing Teletext and Closed Caption data.\n");
if (cap.capabilities & V4L2_CAP_VBI_OUTPUT	) printf("	The device supports the Raw VBI Output interface.\n");
if (cap.capabilities & V4L2_CAP_SLICED_VBI_CAPTURE	) printf("	The device supports the Sliced VBI Capture interface.\n");
if (cap.capabilities & V4L2_CAP_SLICED_VBI_OUTPUT	) printf("	The device supports the Sliced VBI Output interface.\n");
if (cap.capabilities & V4L2_CAP_RDS_CAPTURE	) printf("[to be defined]\n");
if (cap.capabilities & V4L2_CAP_VIDEO_OUTPUT_OVERLAY	) printf("	The device supports the Video Output Overlay (OSD) interface. Unlike the Video Overlay interface, this is a secondary function of video output devices and overlays an image onto an outgoing video signal. When the driver sets this flag, it must clear the V4L2_CAP_VIDEO_OVERLAY flag and vice versa.[a]\n");
if (cap.capabilities & V4L2_CAP_TUNER	) printf("The device has some sort of tuner or modulator to receive or emit RF-modulated video signals. For more information about tuner and modulator programming see Section 1.6.\n");
if (cap.capabilities & V4L2_CAP_AUDIO	) printf("	The device has audio inputs or outputs. It may or may not support audio recording or playback, in PCM or compressed formats. PCM audio support must be implemented as ALSA or OSS interface. For more information on audio inputs and outputs see Section 1.5.\n");
if (cap.capabilities & V4L2_CAP_RADIO) printf("	This is a radio receiver.\n");
if (cap.capabilities & V4L2_CAP_READWRITE) printf("	The device supports the read() and/or write() I/O methods.\n");
if (cap.capabilities & V4L2_CAP_ASYNCIO	) printf("	The device supports the asynchronous I/O methods.\n");
if (cap.capabilities & V4L2_CAP_STREAMING) printf("	The device supports the streaming I/O method.\n");

*/







	switch (io) {
	case IO_METHOD_READ:
		if (!(cap.capabilities & V4L2_CAP_READWRITE)) {
			fprintf (stderr, "%s does not support read i/o\n",
				 dev_name);
			exit (EXIT_FAILURE);
		}

		break;

	case IO_METHOD_MMAP:
	case IO_METHOD_USERPTR:
		if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
			fprintf (stderr, "%s does not support streaming i/o\n",
				 dev_name);
			exit (EXIT_FAILURE);
		}

		break;
	}


        /* Select video input, video standard and tune here. */





	CLEAR (cropcap);

        cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (0 == xioctl (fd, VIDIOC_CROPCAP, &cropcap)) {
                crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                crop.c = cropcap.defrect; /* reset to default */

                if (-1 == xioctl (fd, VIDIOC_S_CROP, &crop)) {
                        switch (errno) {
                        case EINVAL:
                                /* Cropping not supported. */
                                break;
                        default:
                                /* Errors ignored. */
                                break;
                        }
                }
        } else {	
                /* Errors ignored. */
        }




        
        
    video_try_format("V4L2_PIX_FMT_RGB555",V4L2_PIX_FMT_RGB555);
    video_try_format("V4L2_PIX_FMT_RGB565",V4L2_PIX_FMT_RGB565);
    video_try_format("V4L2_PIX_FMT_RGB555X",V4L2_PIX_FMT_RGB555X);
    video_try_format("V4L2_PIX_FMT_RGB565X",      V4L2_PIX_FMT_RGB565X );
    video_try_format("V4L2_PIX_FMT_BGR24",    V4L2_PIX_FMT_BGR24 );
    video_try_format("V4L2_PIX_FMT_RGB24",   V4L2_PIX_FMT_RGB24  );
    video_try_format("V4L2_PIX_FMT_BGR32",  V4L2_PIX_FMT_BGR32);
    video_try_format("V4L2_PIX_FMT_RGB32",    V4L2_PIX_FMT_RGB32 );
    video_try_format("V4L2_PIX_FMT_RGB332",   V4L2_PIX_FMT_RGB332  );
    video_try_format("V4L2_PIX_FMT_YUYV", V4L2_PIX_FMT_YUYV);
    video_try_format("V4L2_PIX_FMT_UYVY",V4L2_PIX_FMT_UYVY);
    video_try_format("V4L2_PIX_FMT_Y41P",    V4L2_PIX_FMT_Y41P   );
    video_try_format("V4L2_PIX_FMT_YVU420",   V4L2_PIX_FMT_YVU420);
    video_try_format("V4L2_PIX_FMT_YUV420",V4L2_PIX_FMT_YUV420);
    video_try_format("V4L2_PIX_FMT_YVU410",  V4L2_PIX_FMT_YVU410);
    video_try_format("V4L2_PIX_FMT_YUV410",V4L2_PIX_FMT_YUV410);
    video_try_format("V4L2_PIX_FMT_YUV422P",V4L2_PIX_FMT_YUV422P);
    video_try_format("V4L2_PIX_FMT_YUV411P", V4L2_PIX_FMT_YUV411P);
    video_try_format("V4L2_PIX_FMT_NV12", V4L2_PIX_FMT_NV12);
    video_try_format("V4L2_PIX_FMT_GREY",   V4L2_PIX_FMT_GREY);
      
      
      /*
    video_try_format("V4L2_PIX_FMT_SBGGR8 ", V4L2_PIX_FMT_SBGGR8);
    video_try_format("V4L2_PIX_FMT_SBGGR16",   V4L2_PIX_FMT_SBGGR16 );
    video_try_format("V4L2_PIX_FMT_GREY ",  V4L2_PIX_FMT_GREY);
    video_try_format("V4L2_PIX_FMT_Y16 ",   V4L2_PIX_FMT_Y16 );
    video_try_format("V4L2_PIX_FMT_YUYV",    V4L2_PIX_FMT_YUYV );


//    video_try_format("V4L2_PIX_FMT_VYUY",  V4L2_PIX_FMT_VYUY);
    
    
    
    video_try_format("V4L2_PIX_FMT_UYVY ",   V4L2_PIX_FMT_UYVY );
    video_try_format("V4L2_PIX_FMT_Y41P ",   V4L2_PIX_FMT_Y41P );
    video_try_format("V4L2_PIX_FMT_YVU420 ",   V4L2_PIX_FMT_YVU420 );
    video_try_format("V4L2_PIX_FMT_YUV420  ",  V4L2_PIX_FMT_YUV420 );
    video_try_format("V4L2_PIX_FMT_YVU410 ",   V4L2_PIX_FMT_YVU410 );
    video_try_format("V4L2_PIX_FMT_YUV410 ",    V4L2_PIX_FMT_YUV410 );
    video_try_format("V4L2_PIX_FMT_YUV422P ",   V4L2_PIX_FMT_YUV422P );
    video_try_format("V4L2_PIX_FMT_YUV411P ",   V4L2_PIX_FMT_YUV411P );
    video_try_format("V4L2_PIX_FMT_NV12 ",  V4L2_PIX_FMT_NV12 );
    video_try_format("V4L2_PIX_FMT_NV21 ",  V4L2_PIX_FMT_NV21);
    video_try_format("V4L2_PIX_FMT_JPEG ",  V4L2_PIX_FMT_JPEG);
    video_try_format("V4L2_PIX_FMT_MPEG ",  V4L2_PIX_FMT_MPEG);
*/


    video_try_format("V4L2_PIX_FMT_MJPEG ",  V4L2_PIX_FMT_MJPEG);






        fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (-1 == xioctl (fd, VIDIOC_G_FMT, &fmt))
                errno_exit ("VIDIOC_G_FMT");



     //   CLEAR (fmt);

        fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        
//        fmt.fmt.pix.width       = 320; 
//        fmt.fmt.pix.height      = 240;

        fmt.fmt.pix.width       = 640; 
       fmt.fmt.pix.height      = 480;




      fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_MJPEG;
    
    
    //    fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;
 //   fmt.fmt.pix.field       = V4L2_FIELD_NONE;

        if (-1 == xioctl (fd, VIDIOC_S_FMT, &fmt))
                errno_exit ("VIDIOC_S_FMT");



  // CLEAR (fmt);
   //   fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (-1 == xioctl (fd, VIDIOC_G_FMT, &fmt))
                errno_exit ("VIDIOC_G_FMT");



     fprintf(stderr,"   fmt.type               %x\n", fmt.type )  ;
     fprintf(stderr,"        fmt.fmt.pix.width    %d\n",    fmt.fmt.pix.width  )  ;
     fprintf(stderr,"        fmt.fmt.pix.height   %d\n",  fmt.fmt.pix.height  )  ;
     fprintf(stderr,"        fmt.fmt.pix.pixelformat %c%c%c%c\n", 
     
             fmt.fmt.pix.pixelformat >> 24, 
             (fmt.fmt.pix.pixelformat >> 16) & 0xff, 
             (fmt.fmt.pix.pixelformat >> 8) & 0xff, 
              fmt.fmt.pix.pixelformat  & 0xff
     
     )  ;



image_x = fmt.fmt.pix.width;
image_y = fmt.fmt.pix.height;


max_image = fmt.fmt.pix.sizeimage;


        /* Note VIDIOC_S_FMT may change width and height. */

	/* Buggy driver paranoia. */
	min = fmt.fmt.pix.width * 2;
	if (fmt.fmt.pix.bytesperline < min)
		fmt.fmt.pix.bytesperline = min;
	min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
	if (fmt.fmt.pix.sizeimage < min)
		fmt.fmt.pix.sizeimage = min;

	switch (io) {
	case IO_METHOD_READ:
		init_read (fmt.fmt.pix.sizeimage);
		break;

	case IO_METHOD_MMAP:
		init_mmap ();
		break;

	case IO_METHOD_USERPTR:
		init_userp (fmt.fmt.pix.sizeimage);
		break;
	}
}



void cameradrv_stop          (void)
{
  
    
           stop_capturing ();

        uninit_device ();


//free(dev_name);





        if (-1 == close (fd))
	        errno_exit ("close");

        fd = -1;
}



int cameradrv_start (char * name, unsigned int *image_x_val, unsigned int *image_y_val, unsigned int *max_image_val)
{

    //dev_name = (char *) malloc(strlen(name) + 2);
        
    
     strcpy(dev_name, name);
//     printf("Copy complete\n");
     
      
     //   dev_name = "/dev/video0";



        struct stat st; 

        if (-1 == stat (dev_name, &st)) {
                fprintf (stderr, "Cannot identify '%s': %d, %s\n",
                         dev_name, errno, strerror (errno));
                exit (EXIT_FAILURE);
        }

        if (!S_ISCHR (st.st_mode)) {
                fprintf (stderr, "%s is no device\n", dev_name);
                exit (EXIT_FAILURE);
        }

        fd = open (dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);

        if (-1 == fd) {
                fprintf (stderr, "Cannot open '%s': %d, %s\n",
                         dev_name, errno, strerror (errno));
                exit (EXIT_FAILURE);
        }
        
        
        init_device ();


*image_x_val = image_x;
*image_y_val = image_y;
*max_image_val = max_image;



        start_capturing ();


        
}




