/*
    Robot server - 2005 Jason Hunt
    
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    Contact: nulluser@gmail.com
    
    File: Camera_drv.c
    
    Adapted from V4L2 reference code
*/

// Project Headers
#include "camera_drv.h"
#include "process.h"

// System Headers
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>         
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <asm/types.h>         
#include <linux/videodev2.h>


#define MODULE_NAME "[Camera Drv] "


/* Extended ioclt */
int xioctl (int fd, int request, void *arg)
{
    int r;

    do 
        r = ioctl (fd, request, arg);
    while (-1 == r && EINTR == errno);

    return r;
}
/* End of xioctl */




int read_frame (camera_type *camera, unsigned char *data, unsigned int max, unsigned int *len)
{
    struct v4l2_buffer buf;

    CLEAR (buf);

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;

    if (xioctl (camera->fd, VIDIOC_DQBUF, &buf) == -1) 
    {
        switch (errno) 
        {
            case EAGAIN: return 0;
			case EIO:
			default:     	error(MODULE_NAME "VIDIOC_DQBUF");
        }
    }

    assert (buf.index < camera->n_buffers);

    //process_image(camera, camera->buffers[buf.index].start, data, buf.bytesused, len);


    *len = buf.bytesused;

    memcpy(data, camera->buffers[buf.index].start, *len);



    if (xioctl (camera->fd, VIDIOC_QBUF, &buf) == -1)
	       	error(MODULE_NAME "VIDIOC_QBUF");

	return 1;
}


void cameradrv_frame (camera_type *camera, unsigned char *data, unsigned int max, unsigned int *len)
{
    for (;;) 
    {
        fd_set fds;
        struct timeval tv;
        int r;

        FD_ZERO (&fds);
        FD_SET (camera->fd, &fds);


        /* Timeout. */
        tv.tv_sec = 2;
        tv.tv_usec = 0;
             
        r = select (camera->fd + 1, &fds, NULL, NULL, &tv);

        if (-1 == r) 
        {
            if (EINTR == errno)
                continue;

                	error(MODULE_NAME "select");
        }

        if (0 == r) 
        {
                	error(MODULE_NAME "select timeout\n");
            return;

        }

		if (read_frame (camera, data, max, len))
     		break;
    }
}









int init_mmap (camera_type *camera)
{
	struct v4l2_requestbuffers req;

    CLEAR (req);

    req.count               = 4;
    req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory              = V4L2_MEMORY_MMAP;

	if (-1 == xioctl (camera->fd, VIDIOC_REQBUFS, &req)) 
    {
        if (EINVAL == errno) 
        {
            error(MODULE_NAME  "%s does not support memory mapping\n", camera->dev_name);
            return 1;         

        } else 
        {
            error(MODULE_NAME "VIDIOC_REQBUFS");
            return 1;
        }
    }

    if (req.count < 2) 
    {
        error(MODULE_NAME "Insufficient buffer memory on %s\n", camera->dev_name);
        return 1;
    }


    camera->buffers = calloc (req.count, sizeof (*camera->buffers));

    if (!camera->buffers) 
    {
        error(MODULE_NAME  "Out of memory\n");
        return 1;
    }

    for (camera->n_buffers = 0; camera->n_buffers < req.count; ++camera->n_buffers) 
    {
        struct v4l2_buffer buf;

        CLEAR (buf);

        buf.type    = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory  = V4L2_MEMORY_MMAP;
        buf.index   = camera->n_buffers;

        if (-1 == xioctl (camera->fd, VIDIOC_QUERYBUF, &buf))
        {
            error(MODULE_NAME "VIDIOC_QUERYBUF");
            return 1;
        }

        camera->buffers[camera->n_buffers].length = buf.length;

        camera->buffers[camera->n_buffers].start =
                        mmap (NULL /* start anywhere */,
                              buf.length,
                              PROT_READ | PROT_WRITE /* required */,
                              MAP_SHARED /* recommended */,
                              camera->fd, buf.m.offset);

        if (MAP_FAILED == camera->buffers[camera->n_buffers].start)
        {
            error(MODULE_NAME "mmap");
            return 1;
        }
    }

    return 0;
    
}




int video_try_format(camera_type *camera, char *name,  unsigned int format)
{
    struct v4l2_format fmt;
    
    CLEAR (fmt);
        
    fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width       = 640; 
    fmt.fmt.pix.height      = 480;
    fmt.fmt.pix.pixelformat = format ;// V4L2_PIX_FMT_YUYV;

    if (-1 == xioctl (camera->fd, VIDIOC_TRY_FMT, &fmt)) 
    {
        // falert(stderr,"Format FAIL %s\n", name);
    }
    else
    {
        alert(MODULE_NAME "Format OK   %s\n", name);
        return 1;
    }

    return 0;
}



void uninit_device  (camera_type *camera )
{
    unsigned int i;

    for (i = 0; i < camera->n_buffers; ++i)
	   if (-1 == munmap (camera->buffers[i].start, camera->buffers[i].length))
	          	error(MODULE_NAME "munmap");
	
	free (camera->buffers);
}



int init_device(camera_type *camera )
{
    log_write(MODULE_NAME "init_device \n");
    
    struct v4l2_capability cap;
    struct v4l2_format fmt;
	unsigned int min;

    if (-1 == xioctl (camera->fd, VIDIOC_QUERYCAP, &cap)) 
    {
        if (EINVAL == errno) 
        {
            error(MODULE_NAME  "%s is no V4L2 device\n", camera->dev_name);
            return 1;
        } 
        else 
        {
            error(MODULE_NAME "VIDIOC_QUERYCAP");
        }
    }

    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) 
    {
           	error(MODULE_NAME  "%s is no video capture device\n",  camera->dev_name);
        return 1;
    }

    if (!(cap.capabilities & V4L2_CAP_STREAMING)) 
    {
            	error(MODULE_NAME   "%s does not support streaming i/o\n", camera->dev_name);
		return 1;
    }


    /* Select video input, video standard and tune here. */

/* 
    struct v4l2_cropcap cropcap;
    struct v4l2_crop crop;


   CLEAR (cropcap);

    cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (0 == xioctl (camera, camera->fd, VIDIOC_CROPCAP, &cropcap)) 
    {
        crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        crop.c = cropcap.defrect; // reset to default 

        if (-1 == xioctl (camera, camera->fd, VIDIOC_S_CROP, &crop)) 
        {
            switch (errno) 
            {
                case EINVAL:
                // Cropping not supported. 
                break;
                default:
                // Errors ignored. 
                 break;
            }
        }
    } else 
    {	
                // Errors ignored. 
    }*/
    
        
 /*   video_try_format(camera, "V4L2_PIX_FMT_RGB555",  V4L2_PIX_FMT_RGB555);
    video_try_format(camera, "V4L2_PIX_FMT_RGB565",  V4L2_PIX_FMT_RGB565);
    video_try_format(camera, "V4L2_PIX_FMT_RGB555X", V4L2_PIX_FMT_RGB555X);
    video_try_format(camera, "V4L2_PIX_FMT_RGB565X", V4L2_PIX_FMT_RGB565X);
    video_try_format(camera, "V4L2_PIX_FMT_BGR24",   V4L2_PIX_FMT_BGR24);
    video_try_format(camera, "V4L2_PIX_FMT_RGB24",   V4L2_PIX_FMT_RGB24);
    video_try_format(camera, "V4L2_PIX_FMT_BGR32",   V4L2_PIX_FMT_BGR32);
    video_try_format(camera, "V4L2_PIX_FMT_RGB32",   V4L2_PIX_FMT_RGB32);
    video_try_format(camera, "V4L2_PIX_FMT_RGB332",  V4L2_PIX_FMT_RGB332);
    video_try_format(camera, "V4L2_PIX_FMT_YUYV",    V4L2_PIX_FMT_YUYV);
    video_try_format(camera, "V4L2_PIX_FMT_UYVY",    V4L2_PIX_FMT_UYVY);
    video_try_format(camera, "V4L2_PIX_FMT_Y41P",    V4L2_PIX_FMT_Y41P);
    video_try_format(camera, "V4L2_PIX_FMT_YVU420",  V4L2_PIX_FMT_YVU420);
    video_try_format(camera, "V4L2_PIX_FMT_YUV420",  V4L2_PIX_FMT_YUV420);
    video_try_format(camera, "V4L2_PIX_FMT_YVU410",  V4L2_PIX_FMT_YVU410);
    video_try_format(camera, "V4L2_PIX_FMT_YUV410",  V4L2_PIX_FMT_YUV410);
    video_try_format(camera, "V4L2_PIX_FMT_YUV422P", V4L2_PIX_FMT_YUV422P);
    video_try_format(camera, "V4L2_PIX_FMT_YUV411P", V4L2_PIX_FMT_YUV411P);
    video_try_format(camera, "V4L2_PIX_FMT_NV12",    V4L2_PIX_FMT_NV12);
    video_try_format(camera, "V4L2_PIX_FMT_NV21",    V4L2_PIX_FMT_NV21);
    video_try_format(camera, "V4L2_PIX_FMT_Y16 ",   V4L2_PIX_FMT_Y16 );
    video_try_format(camera, "V4L2_PIX_FMT_SBGGR8 ", V4L2_PIX_FMT_SBGGR8);
    video_try_format(camera, "V4L2_PIX_FMT_SBGGR16",   V4L2_PIX_FMT_SBGGR16 );
    video_try_format(camera, "V4L2_PIX_FMT_GREY",    V4L2_PIX_FMT_GREY);
    video_try_format(camera, "V4L2_PIX_FMT_JPEG",    V4L2_PIX_FMT_JPEG);
    video_try_format(camera, "V4L2_PIX_FMT_MJPEG ",  V4L2_PIX_FMT_MJPEG);*/
  
  
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (-1 == xioctl (camera->fd, VIDIOC_G_FMT, &fmt))
    {
        error(MODULE_NAME "VIDIOC_G_FMT");
        return 1;
    }

     //   CLEAR (fmt);

    fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    fmt.fmt.pix.width       = camera->req_x; 
    fmt.fmt.pix.height      = camera->req_y;

    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_MJPEG;
    
    
    //    fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;
 //   fmt.fmt.pix.field       = V4L2_FIELD_NONE;

    if (-1 == xioctl (camera->fd, VIDIOC_S_FMT, &fmt))
    {
           	error(MODULE_NAME "VIDIOC_S_FMT");
        return 1;
    }

    if (-1 == xioctl (camera->fd, VIDIOC_G_FMT, &fmt))
    {
        error(MODULE_NAME "VIDIOC_G_FMT");
        return 1;
    }

/*
     alert(MODULE_NAME "fmt.type                %x\n", fmt.type )  ;
     alert(MODULE_NAME "fmt.fmt.pix.width       %d\n", fmt.fmt.pix.width  )  ;
     alert(MODULE_NAME "fmt.fmt.pix.height      %d\n", fmt.fmt.pix.height  )  ;
     alert(MODULE_NAME "fmt.fmt.pix.pixelformat %c%c%c%c\n", 
             fmt.fmt.pix.pixelformat >> 24, 
             (fmt.fmt.pix.pixelformat >> 16) & 0xff, 
             (fmt.fmt.pix.pixelformat >> 8) & 0xff, 
              fmt.fmt.pix.pixelformat  & 0xff) ;*/


    camera->image_x = fmt.fmt.pix.width;
    camera->image_y = fmt.fmt.pix.height;
    camera->max_image = fmt.fmt.pix.sizeimage*2;

/*
   v4l2_streamparm p;

   if (-1 == xioctl (camera, camera->fd,VIDIOC_G_PARM, &p))
   {
          error(MODULE_NAME "VIDIOC_G_PARM");
        return 1;
   
   }
        
 
    	
      if (-1 == xioctl (camera, camera->fd,VIDIOC_S_PARM, &p))
   {
          error(MODULE_NAME "VIDIOC_S_PARM");
        return 1;
   
   }*/
         	


    /* Note VIDIOC_S_FMT may change width and height. */

	/* Buggy driver paranoia. */
	min = fmt.fmt.pix.width * 2;
	if (fmt.fmt.pix.bytesperline < min)
		fmt.fmt.pix.bytesperline = min;

	min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
	if (fmt.fmt.pix.sizeimage < min)
		fmt.fmt.pix.sizeimage = min;


    if (init_mmap (camera))
    {
        return 1;
    }
    
	
	return 0;
}




void stop_capturing (camera_type *camera )
{
    log_write(MODULE_NAME "stop_capturing\n");

    enum v4l2_buf_type type;

	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (-1 == xioctl (camera->fd, VIDIOC_STREAMOFF, &type))
       	error(MODULE_NAME "VIDIOC_STREAMOFF");
}



int start_capturing( camera_type *camera )
{
    log_write(MODULE_NAME "start_capturing\n");

    unsigned int i;
    enum v4l2_buf_type type;

	for (i = 0; i < camera->n_buffers; ++i) 
    {
        struct v4l2_buffer buf;

   		CLEAR (buf);

        buf.type    = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory  = V4L2_MEMORY_MMAP;
        buf.index   = i;

   		if (-1 == xioctl (camera->fd, VIDIOC_QBUF, &buf))
   		{
        	error(MODULE_NAME "VIDIOC_QBUF");
    	   return 1;
    	}

    }
		
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	if (-1 == xioctl (camera->fd, VIDIOC_STREAMON, &type))
	{
    	error(MODULE_NAME "VIDIOC_STREAMON");
    	return 1;
	}
    	
    return 0;
    	
    	
}



void cameradrv_stop (camera_type *camera)
{
    log_write(MODULE_NAME "cameradrv_stop\n");

    stop_capturing (camera);

    uninit_device (camera);

    //free(dev_name);

    if (-1 == close (camera->fd))
            	error(MODULE_NAME "close");

    camera->fd = -1;
}



int cameradrv_start (camera_type *camera)
{
    log_write(MODULE_NAME "cameradrv_start\n");

    camera->fd = -1;
    camera->buffers = NULL;
    camera->n_buffers = 0;
    camera->image_x = 0;
    camera->image_y = 0;
    camera->max_image = 0;

    struct stat st; 

    if (-1 == stat (camera->dev_name, &st)) 
    {
        error(MODULE_NAME "Cannot identify %s", camera->dev_name);
        return 1;
    }

    if (!S_ISCHR (st.st_mode)) 
    {
        error(MODULE_NAME "%s is no device\n", camera->dev_name);
        return 1;
    }
    
    camera->fd = open (camera->dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);

    if (-1 == camera->fd) 
    {
        error(MODULE_NAME "Cannot open %s", camera->dev_name);
        return 1;
    }
        
    if (init_device (camera))
    {
        error(MODULE_NAME "Unable to load camera %s", camera->dev_name);
        return 1;
    }

    if(start_capturing (camera))
    {
        error(MODULE_NAME "Unable to start_capturing  %s", camera->dev_name);    
    
        return 1;
    }
    

    return 0;
}




